
import datetime
import arrow
from . import exceptions
from flask_login import UserMixin
from shaft import (db,
                   utils,
                   send_mail,
                   _)
from shaft.exceptions import ModelError


class AuthRole(db.Model):

    SUPERADMIN = "SUPERADMIN"  # ALL MIGHTY, RESERVED FOR SYS ADMIN
    ADMIN = "ADMIN"  # App/Site admin
    MANAGER = "MANAGER"  # Limited access, but can approve EDITOR Data
    EDITOR = "EDITOR"  # Rights to write, manage, publish own data
    CONTRIBUTOR = "CONTRIBUTOR"  # Rights to only write and read own data
    MODERATOR = "MODERATOR"   # Moderate content
    MEMBER = "MEMBER"  # Just a member

    # BASIC ROLES
    ROLES = [(99, SUPERADMIN),
               (89, ADMIN),
               (59, MANAGER),
               (49, EDITOR),
               (39, CONTRIBUTOR),
               (29, MODERATOR),
               (10, MEMBER)]

    name = db.Column(db.String(75), index=True)
    level = db.Column(db.Integer, index=True)

    @classmethod
    def _syncdb(cls):
        """
        Shaft specific
        To setup some models data after
        :return:
        """
        [cls.new(level=r[0], name=r[1]) for r in cls.ROLES]

    @classmethod
    def new(cls, name, level):
        name = cls.slug_name(name)
        if not cls.get_by_name(name) and not cls.get_by_level(level):
            return cls.create(name=name, level=level)

    @classmethod
    def get_by_name(cls, name):
        name = cls.slug_name(name)
        return cls.query().filter(cls.name == name).first()

    @classmethod
    def get_by_level(cls, level):
        return cls.query().filter(cls.level == level).first()

    @classmethod
    def slug_name(cls, name):
        return utils.slugify(name)

class AuthUser(db.Model, UserMixin):
    role_id = db.Column(db.Integer, db.ForeignKey(AuthRole.id))
    name = db.Column(db.String(255))
    profile_image = db.Column(db.JSONType)
    contact_email = db.Column(db.String(75), index=True)
    registration_method = db.Column(db.String(255))
    active = db.Column(db.Boolean, default=True, index=True)
    last_login = db.Column(db.DateTime)
    last_visited = db.Column(db.DateTime)
    locale = db.Column(db.String(10), default="en")

    # A secret key to sign data per user. If a key is compromised,
    # we can reset this one only and not affect everyone else
    secret_key = db.Column(db.String(250))

    # Relationship to role
    role = db.relationship(AuthRole)


    # ------ FLASK-LOGIN REQUIRED METHODS ----------------------------------
    @property
    def is_active(self):
        return self.active

    # ---------- END FLASK-LOGIN REQUIREMENTS ------------------------------

    @classmethod
    def search_by_name(cls, query, name):
        query = query.filter(db.or_(cls.name.contains(name)))
        return query

    @property
    def email(self):
        return self.login_email

    @property
    def login_email(self):
        user_login = self.get_login("email")
        return user_login.email if user_login else None

    @property
    def login_email_verfied(self):
        user_login = self.get_login("email")
        return True if user_login and user_login.email_verified else False

    @property
    def profile_image_url(self):
        return self.profile_image.get("url")

    def update_profile_image(self, url, object_name, **kwargs):
        """Update the profile image """
        kwargs.update({
            "url": url,
            "object_name": object_name
        })
        self.update(profile_image=kwargs)

    def has_any_roles(self, *roles):
        """
        Check if user has any of the roles requested
        :param roles: tuple of roles string
        :return: bool
        """
        roles = map(utils.slugify, list(roles))

        return True \
            if AuthRole.query()\
            .join(AuthUser)\
            .filter(AuthRole.name.in_(roles))\
            .filter(AuthUser.id == self.id)\
            .count() \
            else False

    def get_login(self, provider="email", social_name=None):
        """
        Return the UserLogin by type
        :param provider: the type of login: email, username
        :param social_name: if type is SOCIAL, it requires the social name, ie: facebook
        :return: AuthUserLogin
        """

        login = self.logins.filter(AuthUserLogin.provider == provider)
        if social_name:
            login = login.filter(AuthUserLogin.social_provider == social_name)
        return login.first()

    def has_login_type(self, provider):
        return True if self.get_login(provider) else False

    def send_email(self, template, **kwargs):
        """
        To send mail to user's .
        Usually for communication purposes
        :param template:
        :param kwargs: custom data to pass
        :return:
        """
        return send_mail(template=template,
                         to=self.contact_email,
                         user=self.email_data,
                         **kwargs)

    @property
    def email_data(self):
        """
        Return some basic user info for email
        :return:
        """
        return {
            "id": self.id,
            "name": self.name,
            "contact_email": self.contact_email
        }

    def set_role(self, role):
        role_ = AuthRole.get_by_name(role.upper())
        if not role_:
            raise ModelError("Invalid user role: '%s'" % role)
        self.update(role=role_)

    def reset_secret_key(self):
        """
        Run this whenever there is a security change in the account.
        ie: password change.
        BTW, AuthUserLogin.change_password, already performs this method
        """
        self.update(secret_key=utils.guid())

    @property
    def secret_salt(self):
        return "USER:%s" % self.id

    def sign_data(self, data, expires_in=None, url_safe=True):
        """
        To safely sign a user data
        :param data: mixed
        :param expires_in: The time for it to expire
        :param url_safe: bool. If true it will allow it to be passed in URL
        :return: str -  the token/signed data
        """
        if url_safe:
            return utils.sign_url_safe(data,
                                       secret_key=self.secret_key,
                                       salt=self.secret_salt,
                                       expires_in=expires_in)
        else:
            return utils.sign_data(data,
                                   secret_key=self.secret_key,
                                   salt=self.secret_salt,
                                   expires_in=expires_in)

    def unsign_data(self, data, url_safe=True):
        """
        Retrieve the signed data. If it is expired, it will throw an exception
        :param data: token/signed data
        :param url_safe: bool. If true it will allow it to be passed in URL
        :return: mixed, the data in its original form
        """
        if url_safe:
            return utils.unsign_url_safe(data,
                                         secret_key=self.secret_key,
                                         salt=self.secret_salt)
        else:
            return utils.unsign_data(data,
                                     secret_key=self.secret_key,
                                     salt=self.secret_salt)

    def signed_data_match(self, data, matched_data, url_safe=True):
        try:
            u_data = self.unsign_data(data, url_safe=url_safe)
            return u_data == matched_data
        except Exception as e:
            return False

class AuthUserLogin(db.Model):
    TYPE_EMAIL = "email"
    TYPE_SOCIAL = "social"
    PROVIDER_FACEBOOK = "facebook"
    PROVIDER_TWITTER = "twitter"
    PROVIDER_EMAIL = "email"

    user_id = db.Column(db.Integer, db.ForeignKey(AuthUser.id))
    provider = db.Column(db.String(50), index=True)
    email = db.Column(db.EmailType, index=True)
    username = db.Column(db.String(250), index=True)
    social_token = db.Column(db.JSONType)
    social_id = db.Column(db.String(255), index=True)
    email_verified = db.Column(db.Boolean, default=False)
    password_hash = db.Column(db.String(255))
    require_password_change = db.Column(db.Boolean, default=False)
    user = db.relationship(AuthUser, backref=db.backref('logins', lazy='dynamic'))

    @classmethod
    def encrypt_password(cls, password):
        return utils.encrypt_string(password)

    @classmethod
    def new(cls,
            email=None,
            username=None,
            password=None,
            provider="email",
            name=None,
            social_token=None,
            social_id=None,
            random_password=True,
            role=None,
            user=None):


        if not user and not name:
            raise exceptions.AuthError(_("Missing User's name"))
        if email and not utils.is_email_valid(email):
            raise exceptions.AuthError(_("Invalid email"))
        if username and not utils.is_username_valid(email):
            raise exceptions.AuthError(_("Invalid username"))

        data = {"provider": provider}
        if email:
            data["email"] = email

        if provider.lower() in ["email", "username"]:
            if not password and random_password:
                password = utils.generate_random_string()
            if not utils.is_password_valid(password):
                raise exceptions.AuthError(_("Password is required"))
            data["password_hash"] = cls.encrypt_password(password)
            data["email_verified"] = False
            if email:
                if cls.get_by_email(email):
                    raise exceptions.AuthError(_("Email exists already"))
            if username:
                if cls.get_by_username(username):
                    raise exceptions.AuthError(_("Email exists already"))
                data["username"] = username
        elif social_token and social_id:
            data["social_id"] = social_id
            data["social_token"] = social_token
        else:
            raise exceptions.AuthError(_("Invalid provider"))

        if not user:
            if not role:
                role = AuthRole.get_by_name("MEMBER")
            if not role:
                raise exceptions.AuthError(_("Missing User's Role"))
            user = AuthUser(name=name,
                            registration_method=data["provider"].lower(),
                            active=True,
                            role=role)
        data["user"] = user
        login = cls.create(**data)
        login.user.reset_secret_key()
        return login


    @classmethod
    def get_by_email(cls, email):
        """
        Return a User by email address
        """
        return cls.query().filter(cls.email == email).first()

    @classmethod
    def get_by_username(cls, username):
        """
        Return a User by email address
        """
        return cls.query().filter(cls.username == username).first()

    @classmethod
    def get_by_social(cls, provider, social_id):
        return cls.query() \
            .filter(cls.provider == provider) \
            .filter(cls.social_id == social_id) \
            .first()

    @property
    def is_email(self):
        return self.provider.lower() == "email"

    @property
    def is_username(self):
        return self.provider.lower() == "username"

    @property
    def is_social(self):
        return self.provider.lower() not in ["email", "username"]

    @property
    def token(self):
        return self.social_token

    def set_email_verified(self, verified=False):
        self.update(email_verified=verified)

    def password_matched(self, password):
        """
        Check if the password matched the hash
        :returns bool:
        """
        return utils.verify_encrypted_string(password, self.password_hash)

    def change_email(self, email):
        """
        Change account email
        :param email:
        :return: the email provided
        """
        if self.provider != self.TYPE_EMAIL:
            raise ModelError("Invalid login type. Must be the type of email to change email")

        if not utils.is_email_valid(email):
            raise exceptions.AuthError("Email address invalid")
        if self.email != email:
            if self.get_by_email(email):
                raise exceptions.AuthError("Email exists already")
            self.update(email=email)
            return True
        return email

    def change_username(self):
        pass

    def change_password(self, password):
        """
        Change the password.
        :param password:
        :return:
        """
        if self.provider != self.TYPE_EMAIL:
            raise ModelError("Invalid login type. Must be the type of email to change password")
        if not utils.is_password_valid(password):
            raise exceptions.AuthError("Invalid password")
        self.update(password_hash=self.encrypt_password(password),
                    require_password_change=False)

        # Whenever the password is changed, reset the secret key to invalidate
        # any tokens in the wild
        self.user.reset_secret_key()

        return password


    @property
    def active(self):
        return self.user.active




import datetime
import uuid
import arrow
import exceptions
from flask_login import fresh_login_required
import logging
import exceptions
from flask_login import (LoginManager,
                         login_user,
                         current_user,
                         fresh_login_required)
from shaft import (Shaft,
                   _,
                   models,
                   page_meta,
                   request,
                   redirect,
                   flash_success,
                   flash_error,
                   flash_info,
                   abort,
                   recaptcha,
                   storage,
                   decorators as shaft_deco,
                   exceptions as shaft_exc,
                   exceptions,
                   utils,
                   paginate,
                   views
                   )


from shaft.contrib.auth import (is_authenticated,
                                not_authenticated,
                                send_password_reset_email,
                                send_email_verification_email,
                                send_registration_welcome_email,
                                session_set_require_password_change,
                                visible_to_managers,
                                get_user_token,
                                login,
                                register,
                                exceptions,
                                change_password,
                                change_login,
                                update_info,
                                __options__ as auth_options,
                                login_manager,
                                decorators as deco
                                )

__options__ = utils.dict_dot({})


def main(**kwargs):

    options = kwargs.get("options", {})
    nav_title_partial = "AuthAccount/nav_title_partial.html"
    nav_kwargs = kwargs.get("nav_menu", {})
    verify_email = options.get("verify_email") or False

    # @shaft_deco.nav_title(
    #     title=nav_kwargs.pop("title", "My Account") or "My Account",
    #     visible=is_authenticated,
    #     css_id=nav_kwargs.pop("css_id", "auth-account-menu"),
    #     css_class=nav_kwargs.pop("css_class", "auth-account-menu"),
    #     align_right=nav_kwargs.pop("align_right", True),
    #     title_partial=nav_kwargs.pop("title_partial", nav_title_partial),
    #     **nav_kwargs)


class Login(Shaft):

    @classmethod
    def _register(cls, app, **kwargs):

        # Late Binding
        __options__.setdefault("login_view", "main.Index:index")
        __options__.setdefault("logout_view", "main.Index:index")
        __options__.setdefault("login_message", "Please log in to access this page.")
        __options__.setdefault("login_message_category", "info")

        # Nav Title
        nav = __options__.get("login.nav", {})
        nav.setdefault("title", None)
        nav.setdefault("order", 100)
        nav.setdefault("position", "right")
        nav.setdefault("title", _("Account"))
        nav["visible"] = not_authenticated
        shaft_deco.nav_title.add(nav.pop("title"), cls, **nav)

        # Route
        kwargs["base_route"] = __options__.get("login.route", "/")

        # Set auth options
        auth_options.update(__options__)

        # Login Manager
        login_manager.login_view = "auth.Login:login"
        login_manager.login_message = __options__.get("login_message")
        login_manager.login_message_category = __options__.get("login_message_category")

        super(cls, cls)._register(app, **kwargs)

    @shaft_deco.nav_title("Login", visible=not_authenticated)
    @shaft_deco.allow_post_get
    @deco.login_not_required
    @deco.require_login_allowed
    @deco.logout_user
    def login(self):
        if request.method == "POST":
            email = request.form.get("email", "").strip()
            password = request.form.get("password", "").strip()
            try:
                if not email or not password:
                    raise shaft_exc.AppError("Email or Password is empty")

                user = login(email=email,
                             password=password,
                             require_verified_email=__options__.get("verify_email"))
                if not user:
                    raise shaft_exc.AppError("Email or Password is invalid")

                userl = user.get_login()
                if userl.require_password_change is True:
                    flash_info("Password change is required")
                    session_set_require_password_change(True)

                    return redirect(views.auth.user_account.Main.change_password)

                return redirect(request.form.get("next") or __options__.get("login_view"))

            except exceptions.VerifyEmailError as ve:
                return redirect(self.login, email=email, v="1")
            except (shaft_exc.AppError, exceptions.AuthError) as ae:
                flash_error(str(ae))
            except Exception as e:
                logging.exception(e)
                flash_error("Unable to login")

            return redirect(self.login, next=request.form.get("next"))

        page_meta("Login")
        return {
            "email": request.args.get("email"),
            "login_url_next": request.args.get("next", ""),
            "allow_register": __options__.get("allow_register"),
            "show_verification_message": True if request.args.get("v") == "1" else False
        }

    @shaft_deco.nav_title("Logout", visible=False)
    @deco.login_not_required
    @shaft_deco.allow_get
    @deco.logout_user
    def logout(self):
        session_set_require_password_change(False)
        return redirect(__options__.get("logout_view") or self.login)

    @shaft_deco.nav_title("Lost Password")
    @shaft_deco.allow_post_get
    @deco.login_not_required
    # @deco.require_login_allowed
    @deco.logout_user
    def lost_password(self):
        page_meta("Lost Password")

        if request.method == "POST":
            email = request.form.get("email")
            user_l = models.AuthUserLogin.get_by_email(email)
            if user_l:
                send_password_reset_email(user=user_l.user, view_class=self)
                flash_success("A new password has been sent to '%s'" % email)
                return redirect(self.login)
            else:
                flash_error("Invalid email address or email doesn't exist")
                return redirect(self.lost_password)

    @shaft_deco.nav_title("Signup", visible=not_authenticated())
    @shaft_deco.allow_post_get
    @deco.login_not_required
    # @deco.require_login_allowed
    # @deco.require_register_allowed
    @shaft_deco.template("auth/login/register.jade")
    @deco.logout_user
    def register(self):
        """ Registration """
        page_meta("Register")

        if request.method == "POST":
            try:
                if not recaptcha.verify():
                    raise shaft_exc.AppError("Invalid Security code")

                name = request.form.get("name", "").strip()
                email = request.form.get("email", "").strip()
                password = request.form.get("password", "").strip()
                password_confirm = request.form.get("password_confirm", "").strip()

                if not name:
                    raise shaft_exc.AppError("Name is required")
                elif not password or password != password_confirm:
                    raise shaft_exc.AppError("Passwords don't match")
                else:
                    user = register(email=email, password=password, name=name)
                    if __options__.get("verify_email"):
                        send_registration_welcome_email(user, view_class=self)
                        flash_success("A welcome email containing a confirmation link has been sent")
                    return redirect(request.form.get("next") or __options__.get("login_view"))
            except (shaft_exc.AppError, exceptions.AuthError) as ex:
                flash_error(str(ex))
            except Exception as e:
                logging.exception(e)
                flash_error("Unable to register")

            return redirect(self.register, next=request.form.get("next"))

        return dict(login_url_next=request.args.get("next", ""), )

    @shaft_deco.nav_title("Reset Password", visible=False)
    @shaft_deco.route("/reset-password/<token>/<action_token>/", methods=["POST", "GET"])
    @deco.login_not_required
    # @deco.require_login_allowed
    @shaft_deco.template("auth/login/reset_password.jade")
    @deco.logout_user
    def reset_password(self, token, action_token):
        """Reset the user password. It was triggered by LOST-PASSWORD """
        try:
            page_meta("Reset Password")

            user = get_user_token(token, "reset-password")
            if not user or not user.signed_data_match(action_token, "reset-password"):
                raise shaft_exc.AppError("Verification Invalid!")

            if request.method == "POST":
                password = request.form.get("password", "").strip()
                password_confirm = request.form.get("password_confirm", "").strip()
                if not password or password != password_confirm:
                    raise exceptions.AuthError("Password is missing or passwords don't match")

                user_login = user.get_login()
                user_login.change_password(password)
                user_login.set_email_verified(True)
                session_set_require_password_change(False)
                flash_success("Password updated successfully!")
                return redirect(__options__.get("login_view") or self.login)

            return {"token": token, "action_token": action_token}

        except (shaft_exc.AppError, exceptions.AuthError) as ex:
            flash_error(str(ex))
        except Exception as e:
            logging.exception(e)
            flash_error("Unable to reset password")
        return redirect(self.login)

    @shaft_deco.nav_title("Confirm Email", visible=False)
    @shaft_deco.allow_post_get
    @deco.login_not_required
    # @deco.require_login_allowed
    @shaft_deco.template("auth/login/request_email_verification.jade")
    @deco.logout_user
    def request_email_verification(self):
        """"""
        if not __options__.get("verify_email"):
            return redirect(self.login)

        if request.method == "POST":
            email = request.form.get("email")
            if email and utils.is_email_valid(email):
                userl = models.AuthUserLogin.get_by_email(email)
                if userl:
                    if not userl.email_verified:
                        send_email_verification_email(userl.user, view_class=self)
                        flash_success("A verification email has been sent to %s" % email)
                    else:
                        flash_success("Your account is already verified")
                    return redirect(self.login, email=email)
            flash_error("Invalid account")
            return redirect(self.request_email_verification, email=email)

        page_meta("Request Email Verification")
        return {
            "email": request.args.get("email"),
        }

    # @deco.require_login_allowed
    @deco.login_not_required
    @shaft_deco.route("/verify-email/<token>/<action_token>/")
    @deco.logout_user
    def verify_email(self, token, action_token):
        """ To verify an email, it accepts a user token and action token.
         Triggered by confirm_email or admin
         """
        try:
            user = get_user_token(token, "verify-email")
            if not user or not user.signed_data_match(action_token, "verify-email"):
                raise shaft_exc.AppError("Verification Invalid!")
            else:
                user.get_login().set_email_verified(True)
                flash_success("Account verified. You can now login")
                return redirect(self.login, email=user.email)
        except Exception as e:
            logging.exception(e)
            flash_error("Verification Failed!")
        return redirect(self.login)

@deco.login_required
class Account(Shaft):

    @classmethod
    def _register(cls, app, **kwargs):

        # Nav
        nav = __options__.get("account.nav", {})
        nav.setdefault("title", None)
        nav.setdefault("order", 100)
        nav["visible"] = is_authenticated
        nav["position"] = "right"
        title = nav.pop("title") or _("My Account")

        # Custom Nav Title
        # Since Title can also be a callback function
        # We can do some customization to display a more fun account menu
        # Specially in the top nav
        custom_nav_title = __options__.get("account.custom_nav_title")
        if __options__.get("account.set_custom_nav_title") and custom_nav_title:
            def custom_nav():
                if is_authenticated():
                    return custom_nav_title.format(USER_NAME=current_user.name,
                                                   USER_EMAIL=current_user.login_email,
                                                   USER_PROFILE_IMAGE_URL=current_user.profile_image_url)
            title = custom_nav

        # Set the nav title
        shaft_deco.nav_title.add(title, cls, **nav)

        # Route
        kwargs["base_route"] = __options__.get("account.route", "/account/")

        super(cls, cls)._register(app, **kwargs)

    def _assert_current_password(self):
        """Assert the password to make sure it matches the current user """

        user_login = current_user.get_login()
        password = request.form.get("current_password")
        if not user_login.password_matched(password):
            raise exceptions.AuthError("Invalid password")

    @shaft_deco.nav_title("Logout", endpoint="auth.Login:logout", order=100)
    def _(self): pass

    @shaft_deco.nav_title("Account Info", order=1)
    def account_info(self):
        page_meta("Account Info")
        return {
            "__USER_ACCOUNT_FORM__": "auth/Account/edit-forms.jade"
        }


    @shaft_deco.allow_post
    def update_login(self):
        """Update the login email"""
        try:
            self._assert_current_password()
            email = request.form.get("email")
            change_login(current_user, email)
            flash_success(_("Login Email changed successfully!"))
        except exceptions.AuthError as ex:
            flash_error(str(ex))
        except Exception as e:
            logging.exception(e)
            flash_error(_("Unable to change email"))
        return redirect(self.account_info)

    @shaft_deco.allow_post
    def update_password(self):
        """Change password """
        try:
            self._assert_current_password()
            password = request.form.get("password", "").strip()
            password_confirm = request.form.get("password_confirm", "").strip()
            change_password(current_user, password, password_confirm)
            flash_success(_("Your password updated successfully!"))
        except exceptions.AuthError as ex:
            flash_error(str(ex))
        except Exception as e:
            logging.exception(e)
            flash_error(_("Unable to update password"))
        return redirect(self.account_info)

    @shaft_deco.allow_post
    def update_info(self):
        """Update basic account info"""
        try:
            name = request.form.get("name", "")
            file = request.files.get("file")
            kwargs = {}
            if file:
                my_photo = storage.upload(file,
                                          name=utils.guid(),
                                          prefix="profile-images/",
                                          allowed_extensions=["jpg", "jpeg", "png", "gif"])
                if my_photo:
                    kwargs["profile_image"] = {
                        "object_name": my_photo.name,
                        "url": my_photo.url,
                        "path": my_photo.path,
                        "provider": my_photo.provider_name
                    }
            kwargs["name"] = name

            update_info(current_user, **kwargs)
            flash_success(_("Info updated successfully!"))
        except exceptions.AuthError as ex:
            flash_error(str(ex))
        except Exception as e:
            logging.exception(e)
            flash_error(_("Unable to update info"))
        return redirect(self.account_info)

    @shaft_deco.nav_title("Setup Login", visible=False)
    @shaft_deco.allow_post_get
    def setup_login(self):
        return
        user_login = current_user.user_login("email")
        if user_login:
            return redirect(self.account_info)

        if request.method == "POST":
            try:
                email = request.form.get("email")
                password = request.form.get("password")
                password2 = request.form.get("password2")

                if not password.strip() or password.strip() != password2.strip():
                    raise exceptions.AuthError("Passwords don't match")
                else:
                    new_login = models.AuthUserLogin.new(login_type="email",
                                                         user_id=current_user.id,
                                                         email=email,
                                                         password=password.strip())
                    if verify_email:
                        send_registration_welcome_email(new_login.user)
                        flash_success("A welcome email containing a confirmation link has been sent to your email")
                        return redirect(self.account_info)
            except exceptions.AuthError as ex:
                flash_error(ex.message)
                return redirect(self.setup_login)
            except Exception as e:
                logging.exception(e)
                flash_error("Unable to setup login")
                return redirect(self)

@deco.login_required
@deco.accepts_manager_roles
class Admin(Shaft):

    @classmethod
    def _register(cls, app, **kwargs):

        # Nav
        nav = __options__.get("admin.nav", {})
        nav.setdefault("title", None)
        nav.setdefault("order", 100)
        nav["visible"] = visible_to_managers
        title = nav.pop("title") or _("User Admin")
        shaft_deco.nav_title.add(title, cls, **nav)

        # Route
        kwargs["base_route"] = __options__.get("admin.route", "/admin/users/")

        super(cls, cls)._register(app, **kwargs)


    def _confirm_password(self):
        user_login = current_user.user_login("email")
        password = request.form.get("confirm-password")
        if not user_login.password_matched(password):
            raise exceptions.AuthError("Invalid password")
        return True

    @classmethod
    def _user_roles_options(cls):
        _r = models.AuthRole.query()\
            .filter(models.AuthRole.level <= current_user.role.level)\
            .order_by(models.AuthRole.level.desc())
        return [(r.id, r.name.upper()) for r in _r]

    @shaft_deco.nav_title("All Users")
    def index(self):

        page_meta("All Users")

        include_deleted = True if request.args.get("include-deleted") == "y" else False
        name = request.args.get("name")
        email = request.args.get("email")
        role = request.args.get("role")
        sorting = request.args.get("sorting", "name__asc")
        users = models.AuthUser.query(include_deleted=include_deleted)
        users = users.join(models.AuthRole).filter(models.AuthRole.level <= current_user.role.level)

        if name:
            users = models.AuthUser.search_by_name(users, name)
        if email:
            users = users.join(models.AuthUserLogin).filter(models.AuthUserLogin.email.contains(email))
        if role:
            users = users.filter(models.AuthUser.role_id == int(role))
        if sorting and "__" in sorting:
            col, dir = sorting.split("__", 2)
            if dir == "asc":
                users = users.order_by(getattr(models.AuthUser, col).asc())
            else:
                users = users.order_by(getattr(models.AuthUser, col).desc())

        users = paginate(users)

        sorting = [("name__asc", "Name ASC"),
                   ("name__desc", "Name DESC"),
                   ("email__asc", "Email ASC"),
                   ("email__desc", "Email DESC"),
                   ("created_at__asc", "Signup ASC"),
                   ("created_at__desc", "Signup DESC"),
                   ("last_login__asc", "Login ASC"),
                   ("last_login__desc", "Login DESC")]

        return dict(user_roles_options=self._user_roles_options(),
                    sorting_options=sorting,
                    users=users,
                    search_query={
                        "include-deleted": request.args.get("include-deleted", "n"),
                        "role": int(request.args.get("role")) if request.args.get("role") else "",
                        "status": request.args.get("status"),
                        "name": request.args.get("name", ""),
                        "email": request.args.get("email", ""),
                        "sorting": request.args.get("sorting")})

    @shaft_deco.nav_title("User Info", visible=False)
    def info(self, id):
        page_meta("User Info")
        user = models.AuthUser.get(id, include_deleted=True)
        if not user:
            abort(404, "User doesn't exist")

        if current_user.role.level < user.role.level:
            abort(403, "Not enough rights to access this user info")

        return {
            "user": user,
            "user_roles_options": self._user_roles_options()
        }

    @shaft_deco.allow_post
    def action(self):
        id = request.form.get("id")
        action = request.form.get("action")

        try:
            user = models.AuthUser.get(id, include_deleted=True)

            if not user:
                abort(404, "User doesn't exist or has been deleted!")
            if current_user.role.level < user.role.level:
                abort(403, "Not enough power level to update this user info")

            if current_user.id != user.id:
                if action == "activate":
                    user.update(active=True)
                    flash_success("User has been ACTIVATED")
                elif action == "deactivate":
                    user.update(active=False)
                    flash_success("User is now DEACTIVATED")
                elif action == "delete":
                    user.delete()
                    flash_success("User has been DELETED")
                elif action == "undelete":
                    user.delete(False)
                    flash_success("User is now RESTORED")

            if action == "info":
                name = request.form.get("name")
                if not name:
                    abort(400, "Missing Name")
                data = {
                    "name": request.form.get("name"),
                }

                if current_user.id != user.id:
                    user_role = request.form.get("user_role")
                    _role = models.AuthRole.get(user_role)
                    if not _role:
                        raise exceptions.AuthError("Invalid ROLE selected")
                    data["role"] = _role

                user.update(**data)
                flash_success("User info updated successfully!")

            elif action == "change-email":
                user_login = user.get_login()
                if not user_login:
                    raise exceptions.AuthError("This account doesn't have an email login type")
                email = request.form.get("email")
                if email != user_login.email:
                    user_login.change_email(email)
                    flash_success("Email '%s' updated successfully!" % email)

            elif action == "change-password":
                user_login = user.get_login()
                if not user_login:
                    raise exceptions.AuthError("This account doesn't have an email login type")
                password = request.form.get("password", "").strip()
                password_confirm = request.form.get("password_confirm", "").strip()
                change_password(user, password, password_confirm)
                flash_success("Password changed successfully!")

            elif action == "email-reset-password":
                send_password_reset_email(user=user)
                flash_success("Password reset was sent to email")

            elif action == "email-account-verification":
                send_email_verification_email(user)
                flash_success("Email verification was sent")

            elif action == "reset-secret-key":
                user.reset_secret_key()
                flash_success("The account's secret key has been changed")

            elif action == "delete-profile-image":
                profile_image_object_name = user.profile_image.get("object_name")
                if profile_image_object_name:
                    obj = storage.get(profile_image_object_name)
                    if obj is not None:
                        obj.delete()
                        user.update_profile_image(url=None, object_name=None)


                flash_success("Profile Image deleted successfully!")

        except exceptions.AuthError as ae:
            flash_error(ae.message)
        return redirect(self.info, id=id)


    @shaft_deco.allow_post
    def create(self):
        try:

            name = request.form.get("name", "").strip()
            email = request.form.get("email", "").strip()
            password = request.form.get("password", "").strip()
            password_confirm = request.form.get("password_confirm", "").strip()

            user_role = request.form.get("user_role")
            role = models.AuthRole.get(user_role)
            if not role:
                raise exceptions.AuthError("Invalid ROLE selected")
            if not name:
                raise shaft_exc.AppError("Name is required")
            elif not password or password != password_confirm:
                raise shaft_exc.AppError("Passwords don't match")

            user = register(email=email,
                             password=password,
                             name=name,
                             role_name=role.name)

            if user:
                flash_success("New account created successfully!")
                return redirect(self.info, id=user.id)
            else:
                raise exceptions.AuthError("Account couldn't be created")
        except exceptions.AuthError as ae:
            flash_error(ae.message)

        return redirect(self.index)

    @shaft_deco.nav_title("User Roles", order=2)
    @shaft_deco.allow_post_get
    def roles(self):
        """
        Only admin and super admin can add/remove roles
        RESTRICTED ROLES CAN'T BE CHANGED
        """
        roles_max_range = 11
        if request.method == "POST":
            try:
                id = request.form.get("id")
                name = request.form.get("name")
                level = request.form.get("level")
                action = request.form.get("action")
                if name and level:
                    level = int(level)
                    name = name.upper()
                    _levels = [r[0] for r in models.AuthRole.ROLES]
                    _names = [r[1] for r in models.AuthRole.ROLES]
                    if level in _levels or name in _names:
                        raise exceptions.AuthError("Can't modify PRIMARY Roles - name: %s, level: %s " % (name, level))
                    else:
                        if id:
                            role = models.AuthRole.get(id)
                            if role:
                                if action == "delete":
                                    role.update(level=0)  # Free the role
                                    role.delete()
                                    flash_success("Role '%s' deleted successfully!" % role.name)
                                elif action == "update":
                                    if role.level != level and models.AuthRole.get_by_level(level):
                                        raise exceptions.AuthError("Role Level '%s' exists already" % level)
                                    elif role.name != models.AuthRole.slug_name(name) and models.AuthRole.get_by_name(name):
                                        raise exceptions.AuthError("Role Name '%s'  exists already" % name)
                                    else:
                                        role.update(name=name, level=level)
                                        flash_success("Role '%s (%s)' updated successfully" % (name, level))
                            else:
                                raise exceptions.AuthError("Role doesn't exist")
                        else:
                            if models.AuthRole.get_by_level(level):
                                raise exceptions.AuthError("New Role Level '%s' exists already" % level)
                            elif models.AuthRole.get_by_name(name):
                                raise exceptions.AuthError( "New Role Name '%s'  exists already" % name)
                            else:
                                models.AuthRole.new(name=name, level=level)
                                flash_success("New Role '%s (%s)' addedd successfully" % (name, level))
            except exceptions.AuthError as ex:
                flash_error("%s" % ex.message)
            return redirect(self.roles)

        page_meta("User Roles")
        roles = models.AuthRole.query().order_by(models.AuthRole.level.desc())

        allocated_levels = [r.level for r in roles]
        #levels_options = [(l, l) for l in range(1, roles_max_range) if l not in allocated_levels]
        levels_options = [(l, l) for l in range(1, roles_max_range)]

        return {
                "roles": roles,
                "levels_options": levels_options
                }


